/*--------------------------------------------------------------------------------------+
|
|   Supplied under applicable software license agreement.
|
|   Copyright (c) 2021 Bentley Systems, Incorporated. All rights reserved.
|
+---------------------------------------------------------------------------------------*/
#pragma once
#include <list>
#include "LevelTypes.h"
#include "DependencyManager.h"
#include "DgnFile.h"
#include "IViewDraw.h"


DGNPLATFORM_TYPEDEFS    (LsDefinition)

BEGIN_BENTLEY_DGNPLATFORM_NAMESPACE

struct DgnAttachment;
struct LevelDefinitionReadNameMgr;
struct LevelTableHandler;
struct LevelTableChangeMergingExtension;
struct LevelEntryHandler;
struct LevelTableEntryChangeMergingExtension;
struct LooseLevel;
struct TxnElementAddOptions;

/*---------------------------------------------------------------------------------**//**
 \addtogroup LevelCache
 A LevelCache provides access to level definitions.

 DgnFile::GetLevelCacheR accesses the level definitions of a file.
 Levels are stored in a level table. A DgnFile has a FileLevelCache to provide access to the information in its level table.

 FileLevelCache::GetLevel accesses the definition of a level.
 Every element in a file is associated with a level. The Elm_hdr.level member of an element
 is the internal LevelId of the level in the file that contains the element. \ref LevelCacheCodeSamples_GetElementLevelDefinition "(Example)"

 LevelHandle accesses the attributes of a level.
 Level attributes tell you if elements on the level should be displayed, what the element by-level symbology is for that level,
 and so on. \ref LevelCacheCodeSamples_GetByLevelColor "(Example)"

 FileLevelCache::CreateLevel allows you to create a new level. \ref LevelCacheCodeSamples_CreateLevel "(Example)"

 EditLevelHandle allows you to modify the attributes of an existing level. You can only get a EditLevelHandle
 from a FileLevelCache to which you have read-write access.

 DgnAttachment::GetAttachmentLevelCache accesses level definition overrides for a reference attachment.
 When one model is referenced into another, the elements in the referenced model refer to levels in the
 file that contains that referenced model. A reference attachment can override the definition of some or all of the referenced levels.
 A DgnAttachmentLevelCache provides access to both levels that were overridden by the attachment and levels that
 are defined by the referenced file.

 FileLevelCache and DgnAttachmentLevelCache are both derived from LevelCache. LevelCache defines
 the basic level look-up and iteration functions that are common to both kinds of level cache. Only FileLevelCache
 provides functions to create or remove levels or to modify level attributes.

 <h3>Collections and Iterators</h3>
 See \ref LevelCacheCodeSamples_Iterate

 <h3>Example Code</h3>
 See \ref LevelCacheCodeSamples

\anchor LevelCache_Usage
 A LevelCache collaborates with DgnFile and DgnModel to keep track of which levels are actually used by elements. A level is used if:
 \li A graphic element (including element of a shared cell definition) is on the level,
 \li Individual segments of a Multi-line element are on the level,
 \li Individual segments of a Multi-line Style are on the level,
 \li A reference attachment is on the level,
 \li A dimension style has its "Dimension Line Level Override" on the level.

\internal
 \dot
  digraph LevelCacheTypes {
        DgnFile -> FileLevelCache;
        FileLevelCache -> LevelHandle;
        DgnFile -> DgnModel;
        DgnModel -> DgnAttachment;
        DgnAttachment -> DgnAttachmentLevelCache;
        DgnAttachmentLevelCache -> LevelDefinitionHandle_Override;
        DgnAttachmentLevelCache -> FileLevelCache_DgnAttachment;
        LevelDefinitionHandle_Override -> DgnFile_DgnAttachment;
        LevelDefinitionHandle_Override -> LevelDefinitionHandle_DgnAttachment;
        DgnAttachment -> DgnCache_DgnAttachment;
        DgnCache_DgnAttachment -> DgnFile_DgnAttachment;
        DgnFile_DgnAttachment -> FileLevelCache_DgnAttachment;
        FileLevelCache_DgnAttachment -> LevelDefinitionHandle_DgnAttachment;
  }
  \enddot
+---------------+---------------+---------------+---------------+---------------+------*/
//! @{

typedef bset<byte>::iterator T_LevelIdToDefinitionMapIterator;


//=======================================================================================
//! Error codes returned by LevelCache function
// @bsiclass                                                     Bentley Systems

//  NOTE: These originated from MSGID_errorXX in leveltableids.h in 08.11.09
//=======================================================================================
enum class LevelCacheErrorCode
{
    None                            = BSISUCCESS,   //!< Success.
    Error                           = BSIERROR,     //!< An unspecified error.
    TableIsReadOnly                 =  2,           //!< The level cache is marked as read-only.
    LevelIsReadOnly                 =  4,           //!< The level is read-only.
    LevelNameIsNull                 =  7,           //!< The level name cannot be NULL.
    LevelNameIsInvalid              =  8,           //!< The name contains invalid characters. See LevelUtils::ContainsInvalidCharacters.
    LevelNameIsDuplicate            =  9,           //!< The name is already used by a level in this cache.
    LevelIsUsed                     = 11,           //!< The level is in use.
    LevelCodeIsDuplicate            = 23,           //!< The LevelCode is already used by a level in this cache.
    LevelIdIsDuplicate              = 46,           //!< The LevelId is already used by a level in this cache.
    LibraryLevelNotEditable         = 49,           //!< The level is in a library, and cannot be edited.
    CannotFindLevel                 = 56,           //!< Cannot find level.
    LevelTableIntegrityFailure      = 59,           //!< Integrity check failure.
    DgnAttachmentNotResolved        = 60,           //!< Attempt to query a FileLevelCache for a DgnAttachment that is not resolved (not found or not read).
    LevelNullId                     = 92,           //!< A level element with levelId - LEVEL_NULL_ID is in the level table element descriptor and that level definition is ignored.
    NoLevels                        = 1000,         //!< There are no levels in the LevelCache.
    AttachmentSharesFileLevelCache  = 1001,         //!< An attempt was made to Write the LevelCache associated with a DgnAttachment that shares the FileLevelCache for the root model.
};

/**
 * Level Display Name Format Defines
 */
#define LEVEL_NAME_DISPLAY_FORMAT_STRING            L"N"
#define LEVEL_CODE_DISPLAY_FORMAT_STRING            L"C"
#define LEVEL_DESCRIPTION_DISPLAY_FORMAT_STRING     L"D"

/**
 * Level Display Name Format Values
 */
#define LEVEL_NAME_DISPLAY_FORMAT                   L'N'
#define LEVEL_CODE_DISPLAY_FORMAT                   L'C'
#define LEVEL_DESCRIPTION_DISPLAY_FORMAT            L'D'
#define LEVEL_ID_DISPLAY_FORMAT                     L'I'


//=======================================================================================
//! Base class for style that is used by a level.
//! The style refers to data in the definition file.
//  @remark Required library : DgnPlatform<ApiNumber>.lib  i.e. DgnPlatform5.lib
// @bsiclass                                                     Bentley Systems

//=======================================================================================
struct LevelDefinitionStyleBase
{
protected:
    DgnFileP  m_definitionFile;

public:
    DGNPLATFORM_EXPORT  LevelDefinitionStyleBase                (DgnFileP file);
    DGNPLATFORM_EXPORT  LevelDefinitionStyleBase&   operator=   (LevelDefinitionStyleBase const& source);

    //! Get the file that defines a color or style.
    DGNPLATFORM_EXPORT  DgnFileP    GetDefinitionFile       () const;

    //! Get a representative model in that file that defines a color or style.
    DGNPLATFORM_EXPORT  DgnModelP   GetDefinitionModel      () const;
}; // LevelDefinitionStyleBase

//=======================================================================================
//! Represents a color used by a level.
//! A color may be an index color, a true color, or color book entry.
//! The color code refers to an entry in a color table or color book.
//! The definition file contains the table or color book.
//!
//! Example: \ref LevelCacheCodeSamples_GetByLevelColor
//  @remark Required library : DgnPlatform<ApiNumber>.lib  i.e. DgnPlatform5.lib
// @bsiclass                                                     Bentley Systems

//=======================================================================================
struct LevelDefinitionColor : LevelDefinitionStyleBase
{
    DEFINE_T_SUPER(LevelDefinitionStyleBase);
private:
    UInt32      m_color;

public:
    //!  Constructor for LevelDefinitionColor
    //!  @DotNetMethodCustomImplementation
    DGNPLATFORM_EXPORT  LevelDefinitionColor                (UInt32 c, DgnFileP file);
    DGNPLATFORM_EXPORT  LevelDefinitionColor&   operator=   (LevelDefinitionColor const& r);

    //! Get the color code. @See GetDefinitionFile
    DGNPLATFORM_EXPORT UInt32      GetColor () const;

    //! Compare this color for equivalence to another color. The other color can be from a different file.
    DGNPLATFORM_EXPORT bool     operator==  (LevelDefinitionColor const&) const;

    //! Compare this color for equivalence to another color. The other color can be from a different file.
    DGNPLATFORM_EXPORT bool     operator!=  (LevelDefinitionColor const& rhs) const;
}; // LevelDefinitionColor

//=======================================================================================
//! Represents a line style used by a level.
//! The style code refers to a style defined in the definition file.
//  @remark Required library : DgnPlatform<ApiNumber>.lib  i.e. DgnPlatform5.lib
// @bsiclass                                                     Bentley Systems

//=======================================================================================
struct LevelDefinitionLineStyle : LevelDefinitionStyleBase
{
    DEFINE_T_SUPER(LevelDefinitionStyleBase)
private:
    int                 m_style;
    LineStyleParamsCP   m_styleParams;

public:
    //!  Constructor for LevelDefinitionLineStyle
    //!  @DotNetMethodCustomImplementation
    DGNPLATFORM_EXPORT LevelDefinitionLineStyle                 (int c, LineStyleParamsCP p, DgnFileP file);
    DGNPLATFORM_EXPORT LevelDefinitionLineStyle& operator=      (LevelDefinitionLineStyle const& r);

    //! Get the line style code. @See GetDefinitionFile
    DGNPLATFORM_EXPORT int      GetStyle () const;

    //! Compare this line style for equivalence to another line style. The other line style can be from a different file.
    DGNPLATFORM_EXPORT bool operator==      (LevelDefinitionLineStyle const&) const;
    //! Compare this line style for equivalence to another line style. The other line style can be from a different file.
    DGNPLATFORM_EXPORT bool operator!=      (LevelDefinitionLineStyle const& rhs) const;

    //! Query if this definition includes parameters to be used with this line style.
    DGNPLATFORM_EXPORT bool     HasStyleParams () const;

    //! Get the parameters to be used with this line style.
    DGNPLATFORM_EXPORT LineStyleParamsCP GetStyleParamsP () const;

    //! Get the parameters to be used with this line style.
    //! @return false if this definition does not include parameters to be used with this line style.
    DGNPLATFORM_EXPORT bool     GetStyleParams (LineStyleParamsR sp) const;


}; // LevelDefinitionLineStyle

//=======================================================================================
//! Represents a style used by a level. The style is identified by an ElementId.
//! The ElementId refers to an element in the definition file.
//  @remark Required library : DgnPlatform<ApiNumber>.lib  i.e. DgnPlatform5.lib
// @bsiclass                                                     Bentley Systems

//=======================================================================================
struct LevelDefinitionIdBase : LevelDefinitionStyleBase
{
    DEFINE_T_SUPER(LevelDefinitionStyleBase)
protected:
    ElementId   m_id;

public:
    //! Constructor for LevelDefinitionIdBase
    //!  @DotNetMethodCustomImplementation
    DGNPLATFORM_EXPORT LevelDefinitionIdBase (ElementId m, DgnFileP file);
    DGNPLATFORM_EXPORT LevelDefinitionIdBase& operator= (LevelDefinitionIdBase const& r);

    DGNPLATFORM_EXPORT bool operator==      (LevelDefinitionIdBase const&) const;

    //! Get the ElementId that identifies the resource
    DGNPLATFORM_EXPORT ElementId GetId () const;
};

//=======================================================================================
//! Represents a material used by a level.
//  @remark Required library : DgnPlatform<ApiNumber>.lib  i.e. DgnPlatform5.lib
// @bsiclass                                                     Bentley Systems

struct LevelDefinitionMaterial : LevelDefinitionIdBase
{
    DEFINE_T_SUPER(LevelDefinitionIdBase)

    //!  Constructor for LevelDefinitionMaterial
    //!  @DotNetMethodCustomImplementation
    DGNPLATFORM_EXPORT LevelDefinitionMaterial (ElementId m, DgnFileP file);
    DGNPLATFORM_EXPORT LevelDefinitionMaterial& operator= (LevelDefinitionMaterial const& r);

    //! Compare this material for equivalence to another material. The other material can be from a different file.
    DGNPLATFORM_EXPORT bool operator==  (LevelDefinitionMaterial const&) const;

    DGNPLATFORM_EXPORT bool operator!=  (LevelDefinitionMaterial const& rhs) const;

};

//=======================================================================================
//! Represents a PlotStyleID used by a level.
//  @remark Required library : DgnPlatform<ApiNumber>.lib  i.e. DgnPlatform5.lib
// @bsiclass                                                     Bentley Systems

//=======================================================================================
struct LevelDefinitionPlotStyle : LevelDefinitionIdBase
{
    DEFINE_T_SUPER(LevelDefinitionIdBase)

    //!  Constructor for LevelDefinitionPlotStyle
    //!  @DotNetMethodCustomImplementation
    DGNPLATFORM_EXPORT LevelDefinitionPlotStyle (ElementId m, DgnFileP file);
    DGNPLATFORM_EXPORT LevelDefinitionPlotStyle& operator= (LevelDefinitionPlotStyle const& r);

    //! Compare this PlotStyleID for equivalence to another PlotStyleID. The other PlotStyleID can be from a different file.
    DGNPLATFORM_EXPORT bool operator==  (LevelDefinitionPlotStyle const&) const;

    DGNPLATFORM_EXPORT bool operator!=  (LevelDefinitionPlotStyle const& rhs) const;

};

/*=================================================================================**//**
  Read-only access to a level definition.
  Example: \ref LevelCacheCodeSamples_GetElementLevelDefinition

  A LevelHandle is also an iterator over level definitions in a LevelCache.

  Example: \ref LevelCacheCodeSamples_Iterate

  <H1>Level Attributes</H1>
  Each level has the following properties:
  <UL>
  <LI><b>Name</b> - The level name is a Unicode String.
  <LI><b>Code (or Number)</b> - The level code is an unsigned long. A user may or may-not assign a level code.
  All user-assigned numbers must be unique across levels in a level table. The level code can be changed using EditLevelHandle::SetLevelCode.
  To clear a user-assigned level code, supply LEVEL_NULL_CODE to EditLevelHandle::SetLevelCode. User assigned Level code is not supported
  for DWG Files.
  <br>
  The level code should not be confused with the level id, which also an unsigned long that is also
  unique in the context of all levels in a level table. The following table which compares level id with level code:

  <TABLE BORDER>
  <TR> <TH></TH>                    <TH>level id</TH>                               <TH>level code</TH>                         </TR>
  <TR> <TH>Type</TH>                <TD>UInt32</TD>                                 <TD>UInt32</TD>                             </TR>
  <TR> <TH>Uniqueness</TH>          <TD>Unique in context of level-table</TD>       <TD>Unique in context of level-table</TD>   </TR>
  <TR> <TH>Assigned By</TH>         <TD>System</TD>                                 <TD>User</TD>                               </TR>
  <TR> <TH>Is Required</TH>         <TD>Yes</TD>                                    <TD>No, if not assigned by user then system will auto-generate a number</TD>   </TR>
  <TR> <TH>Changeable?</TH>         <TD>No</TD>                                     <TD>Yes</TD>                                </TR>
  <TR> <TH>Stored in elements?</TH> <TD>Yes</TD>                                    <TD>No</TD>                                 </TR>
  </TABLE>

  <LI><b>ByLevel Symbology</b> - Color, Weight and Style. A level's element symbology is assigned to all those elements
  of the level that have their respective symbology set to BYLEVEL. Thus if an element has its color set to COLOR_BYLEVEL,
  then its level's "Element Color" is used to color the element.

  <LI><b>Override Symbology</b> - Color, Weight and Style with a set of ON/OFF Override Flags. A level's symbology
  is assigned to its elements in a particular view, if the "level symbology" attribute of the view is turned on and if the corresponding
  override flag in the level is turned ON. Thus, if a level's override color is set to RED, and if the level color override is ON and if "level symbology"
  is turned on in View 1, then all elements of the level will be displayed in RED.

  <LI><b>Level Access Controls</b> : A level can set to hidden or it can be set to readonly.
  <br>
        <b>Set level hidden on</b>: Making a level hidden makes it hidden from the MicroStation user interface. This can be used by
        third party programs to create levels for their purposes and which the level manager does not display. Beyond that, a hidden
        level is no different from any other level. Call EditLevelHandle::SetHidden to turn on/off the hidden nature of a level.
  <br>
        <b>Set level readonly</b>: This is similar to making a level hidden, except that it is displayed in the MicroStation user interface.
        For example, the level manager will list this level, but its attributes cannot be changed. Exception: The level display controls
        of a readonly level can be changed. Call FileLevelCache::SetLevelReadOnly to turn on/off the readonly nature of a level.
  <br>
        <b>Set level table readonly</b>: The entire level table may be set read-only. Setting a level-table as read-only has
        the following implications:
        <UL>
        <LI>No new levels can be created in the level-table.
        <LI>None of the exisiting levels can be deleted from the level-table.
        <LI>None of the level attributes can be modified.
        <LI>None of the existing level libraries can be detached from the level-table.
        </UL>

  <LI><b>Level Element Access Controls</b> : The access to elements on a level can be controlled by setting a element access mode on
  the level. The element access mode can be set to one of 4 values -
  <TT>LevelElementAccess::All, LevelElementAccess::ReadOnly, LevelElementAccess::ViewOnly, or LevelElementAccess::Locked</TT>. See LevelHandle::GetElementAccess.
  </UL>

 \internal
  The purpose of this "handle" class is to define an API while hiding the LevelDefinition
  storage mechanism. This indirection allows us to change the storage mechanism without breaking the API.
  This handle object also gives us a place to store transient data that is used to create
  an efficient and convenient API. For example, the handle stores a back-pointer to the level's
  owning LevelCache. This back-pointer is necessary for accessing some level data. We don't want to have
  to store this backpointer on a LevelDefinition, however. It is redundant, since a LevelDefinition is
  always contained in and owned by a LevelCache. Likewise, we can store data in this
  temporary handle that supports copy-on-write semantics (see EditLevelHandle).
   @remark Required library : DgnPlatform<ApiNumber>.lib  i.e. DgnPlatform5.lib
  @bsiclass                                                     Bentley Systems

+===============+===============+===============+===============+===============+======*/
struct          LevelHandle : std::iterator<std::forward_iterator_tag, LevelHandle const>
{
private:
protected:
    LevelCacheErrorCode                 m_status;
    T_LevelIdToDefinitionMapIterator    m_levelIterator;
    LevelCacheP                         m_levelCache;
public:


/// \name Constructors
/// @{

//! Create an empty, invalid handle
DGNPLATFORM_EXPORT  LevelHandle ();
//! Duplicate a handle
DGNPLATFORM_EXPORT  LevelHandle (LevelHandle const&);
//! Duplicate a handle
DGNPLATFORM_EXPORT  LevelHandle& operator= (LevelHandle const&);

//! @}

//! \name Identifiers and Status
//! @{

//! Get the internal identifier for this level
DGNPLATFORM_EXPORT LevelId GetLevelId () const;

//! Get the description or purpose of this level. May be NULL.
DGNPLATFORM_EXPORT WCharCP GetDescription () const;

//! Get the name of this level. May be NULL.
DGNPLATFORM_EXPORT WCharCP GetName () const;

//! Get the display name of this level.
DGNPLATFORM_EXPORT BentleyStatus GetDisplayName (WStringR displayNameStr) const;

//! Format the display name of this level from a format string.
//! This method is generally used by the _FormatLevelDisplayName method of implementations of LevelCacheAdmin.
DGNPLATFORM_EXPORT BentleyStatus FormatDisplayName (WStringR displayNameStr, WCharCP formatString) const;

//! Get the optional code assigned to this level
DGNPLATFORM_EXPORT LevelCode GetLevelCode () const;

//! Get the level cache that contains this level
DGNPLATFORM_EXPORT LevelCache* GetLevelCacheP () const;

//! Returns true if this handle refers to a level, false if this is an invalid handle.
//! Example: \ref LevelCacheCodeSamples_CreateLevel
//! @see GetStatus to query the reason why a handle is invalid.
DGNPLATFORM_EXPORT bool IsValid () const;

//! Is this level an alias for another level in another level cache?
DGNPLATFORM_EXPORT bool IsAlias () const;


//! Returns non-zero error status if this is an invalid handle. The status indicates why the level cache operation that produced this handle failed.
//! Example: \ref LevelCacheCodeSamples_CreateLevel
//! @see IsValid
DGNPLATFORM_EXPORT LevelCacheErrorCode GetStatus () const;

//! @}

//! \name Comparison
//! Example: \ref LevelCacheCodeSamples_Iterate
/// @{

//! Test if this handle refers to the very same level as another handle
DGNPLATFORM_EXPORT bool operator== (LevelHandle const&) const;
//! Test if this handle does not refer to the very same level as another handle
DGNPLATFORM_EXPORT bool operator!= (LevelHandle const&) const;

//! Compare the data contained this level definition to the data contained in \a sourceLevel.
//! @returns true if the specified attributes match.
//! @param          differences         [optional] if not NULL, records the attributes that do not match.
//! @param          sourceLevel         The level to which this level is to be compared.
//! @param          attributesToCompare [optional] if not NULL, specifies which attributes to compare. If null, all attributes are compared.
//! @remarks When comparing levels from different files, EqualLevelData attempts to detect if colors and styles are \em equivalent.
//! @remarks EqualLevelData never compares LevelIDs. If you want to detect if two levels have the same LevelId, you must call GetLevelId on each of them.
//!             Two levels in the same level cache will never have the same LevelId. It is rarely meaningful to compare the LevelIDs of levels from different caches.
//! @remarks This function does not compare the ElementIds of the levels
DGNPLATFORM_EXPORT bool EqualLevelData (BitMaskP differences, LevelHandle const& sourceLevel, BitMaskCP attributesToCompare) const;
//! @}

/// \name Iteration
//! Example: \ref LevelCacheCodeSamples_Iterate
/// @{
DGNPLATFORM_EXPORT LevelHandle& operator++ ();
/// @}

DGNPLATFORM_EXPORT  LevelHandle const& operator*  () const;
DGNPLATFORM_EXPORT  LevelHandle const* operator-> () const;

/// \name Attributes
/// @{

//! Get the color to use when applying level symbology
DGNPLATFORM_EXPORT LevelDefinitionColor GetOverrideColor () const;
//! Get the style to use when applying level symbology
DGNPLATFORM_EXPORT LevelDefinitionLineStyle GetOverrideLineStyle () const;
//! Get the weight to use when applying level symbology
DGNPLATFORM_EXPORT UInt GetOverrideWeight () const;
//! Get the material to use when applying level symbology
DGNPLATFORM_EXPORT LevelDefinitionMaterial GetOverrideMaterial () const;
//! Get the color to use when an element's color is defined as "by-level"
//! Example: \ref LevelCacheCodeSamples_GetByLevelColor
DGNPLATFORM_EXPORT LevelDefinitionColor GetByLevelColor () const;
//! Get the style to use when an element's style is defined as "by-level"
DGNPLATFORM_EXPORT LevelDefinitionLineStyle GetByLevelLineStyle () const;
//! Get the weight to use when an element's weight is defined as "by-level"
DGNPLATFORM_EXPORT UInt GetByLevelWeight () const;
//! Get the material to use when an element's material is defined as "by-level"
DGNPLATFORM_EXPORT LevelDefinitionMaterial GetByLevelMaterial () const;
//! Query if level color should be used when applying level symbology
DGNPLATFORM_EXPORT bool GetOverrideColorOn () const;
//! Query if level style should be used when applying level symbology
DGNPLATFORM_EXPORT bool GetOverrideLineStyleOn () const;
//! Query if level weight should be used when applying level symbology
DGNPLATFORM_EXPORT bool GetOverrideWeightOn () const;
//! Query if level material should be used when applying level symbology
DGNPLATFORM_EXPORT bool GetOverrideMaterialOn () const;
//! Query if elements on this level should be displayed
DGNPLATFORM_EXPORT bool GetDisplay () const;
//! Query if elements on this level should be plotted
DGNPLATFORM_EXPORT bool GetPlot () const;
//! \em TBD...
DGNPLATFORM_EXPORT bool GetByCell () const;
//! Query if elements on this level may be snapped to
DGNPLATFORM_EXPORT bool GetSnap () const;
//! Query if this level should not be modified
DGNPLATFORM_EXPORT bool GetReadOnly () const;
//! Query if this level is hidden, i.e., should not be shown to the user
DGNPLATFORM_EXPORT bool GetHidden () const;
//! Query access restrictions to be applied to elements on this level.
//! @returns
//! -- LevelElementAccess::All: Elements on the level have all access. This is the default mode
//! -- LevelElementAccess::Locked: Elements on the level are locked. They can be copied, but cannot be modified or deleted. New elements can be added to the level - once added they are immediately locked. The level can be set active.
DGNPLATFORM_EXPORT LevelElementAccess GetElementAccess () const;
//! Query if elements on this level are frozen
DGNPLATFORM_EXPORT bool GetFrozen () const;
//! Query if elements on this level may be located
DGNPLATFORM_EXPORT bool GetLocate () const;
//! Query if the LevelCode associated with this level was generated and assigned by the LevelCache automatically, rather than by the user.
DGNPLATFORM_EXPORT bool GetCodeAutoGenerated () const;
//! Get the level display priority
DGNPLATFORM_EXPORT Int32 GetDisplayPriority () const;
//! Get the level display transparency
DGNPLATFORM_EXPORT double GetTransparency () const;
//! Get the PlotStyle
DGNPLATFORM_EXPORT LevelDefinitionPlotStyle GetPlotStyle () const;

/// @}

}; // LevelHandle

//=======================================================================================
//!
//! Read-write access to a level definition
//!
//! Collaborates with LevelCache to track changes to level definitions.
//! Collaborates with ReferenceAttachmentLevelCache to implement copy-on-write.
//!
//! Note that you cannot create a EditLevelHandle for a level that is
//! read-only or is in a level cache that is marked read-only.
//  @remark Required library : DgnPlatform<ApiNumber>.lib  i.e. DgnPlatform5.lib
// @bsiclass                                                     Bentley Systems

//=======================================================================================
struct          EditLevelHandle : public LevelHandle
{
    DEFINE_T_SUPER(LevelHandle)
private:
public:


//! \name Constructors
//! @{

//! Create an empty, invalid handle
DGNPLATFORM_EXPORT EditLevelHandle ();

//! Copy a handle
DGNPLATFORM_EXPORT EditLevelHandle (EditLevelHandle const&);

//! Get read-write access to a level. This will fail if the source level is read-only.
DGNPLATFORM_EXPORT explicit EditLevelHandle (LevelHandle const&);


//! Query if this level has been changed since the last call to FileLevelCache::Write
DGNPLATFORM_EXPORT bool IsDirty () const;

//! @}

//! \name Identifiers
//! @{

//! Change the name of this level.
//! @returns
//!     \li LevelCacheErrorCode::TableIsReadOnly - The level table is read-only
//!     \li LevelCacheErrorCode::LevelNameIsDuplicate - \a name is a name that is already used by another level in this cache
//!     \li LevelCacheErrorCode::LevelNameIsInvalid - \a name contains invalid characters. See LevelUtils::ContainsInvalidCharacters
DGNPLATFORM_EXPORT LevelCacheErrorCode SetName (WCharCP name);

//! Change the LevelCode of this level.
//! @param[in] levelCode The level code to set.
//! @param[in] autoGenerated If true, then the system does not attempt to retain this Level Code. If false, the system retains that level code from library levels.
//! @returns
//!     \li LevelCacheErrorCode::TableIsReadOnly - The level table is read-only
//!     \li LevelCacheErrorCode::LevelCodeIsDuplicate - \a code is a LevelCode that is already used by another level in this cache
DGNPLATFORM_EXPORT LevelCacheErrorCode SetLevelCode (LevelCode levelCode, bool autoGenerated);

//! Change the description of this level.
//! @returns true if the new description is different from the old.
DGNPLATFORM_EXPORT bool SetDescription (WCharCP);

//! @}

//! \name Copying
//! @{

//! Copy a handle. The copy refers to the very same level as the original.
DGNPLATFORM_EXPORT EditLevelHandle& operator= (EditLevelHandle const&);

//! Copy attribute values from \a sourceLevel to this level.
//! @remarks LEVEL_ID_ATTRIBUTE, LEVEL_CODE_ATTRIBUTE, and LEVEL_NAME_ATTRIBUTE are not supported.
//! In order to change a level's code or name, you must call SetLevelCode or SetName. These methods
//! can fail. You cannot change a level's ID. Call CreateLevel in order to create a level with a specified LevelId.
//! @param sourceLevel The level to copy from.
//! @param m [optional] Identifies the properties to copy.
//! @return true if any property value was actually modified
DGNPLATFORM_EXPORT bool CopyLevelData (LevelHandle const& sourceLevel, BitMaskCP m=NULL);

//! @}

//! \name Attributes
//! @{

//! Set the color to use when applying level symbology. @returns true if the new value is different from the current value.
DGNPLATFORM_EXPORT bool SetOverrideColor (LevelDefinitionColor const&);
//! Set the style to use when applying level symbology.
//! @param l    A valid line style definition
//! @param p    Optional style params
//! @param f    The file in which the line style is defined
//! @returns true if the new line style code is different from the current value.
//! @remarks The level records only the style number. This number refers to the specified style definition in the definition file (or in a resource file that the host opens).
DGNPLATFORM_EXPORT bool SetOverrideLineStyle (LsEntryCR l, LineStyleParamsCP p, DgnFileR f);
//! Set the weight to use when applying level symbology. @returns true if the new value is different from the current value.
DGNPLATFORM_EXPORT bool SetOverrideWeight (UInt);
//! Set the material to use when applying level symbology. @returns true if the new value is different from the current value.
DGNPLATFORM_EXPORT bool SetOverrideMaterial (LevelDefinitionMaterial const&);
//! Set the color to use when an element's color is defined as "by-level". @returns true if the new value is different from the current value.
DGNPLATFORM_EXPORT bool SetByLevelColor (LevelDefinitionColor const&);

DGNPLATFORM_EXPORT bool SetByLevelLineStyle (LevelDefinitionLineStyle const&);  //Added by Foo [2022-01-27]

//! Set the style to use when an element's style is defined as "by-level". @returns true if the new value is different from the current value.
//! @param l    A valid line style definition
//! @param p    Optional style params
//! @param f    The file in which the line style is defined
//! @returns true if the new line style code is different from the current value.
//! @remarks The level records only the style number. This number refers to the specified style definition in the definition file (or in a resource file that the host opens).
DGNPLATFORM_EXPORT bool SetByLevelLineStyle (LsEntryCR l, LineStyleParamsCP p, DgnFileR f);
//! Set the weight to use when an element's weight is defined as "by-level". @returns true if the new value is different from the current value.
DGNPLATFORM_EXPORT bool SetByLevelWeight (UInt);
//! Set the material to use when an element's material is defined as "by-level". @returns true if the new value is different from the current value.
DGNPLATFORM_EXPORT bool SetByLevelMaterial (LevelDefinitionMaterial const&);
//! Specify if level color should be used when applying level symbology. @returns true if the new value is different from the current value.
DGNPLATFORM_EXPORT bool SetOverrideColorOn (bool);
//! Specify if level style should be used when applying level symbology. @returns true if the new value is different from the current value.
DGNPLATFORM_EXPORT bool SetOverrideLineStyleOn (bool);
//! Specify if level weight should be used when applying level symbology. @returns true if the new value is different from the current value.
DGNPLATFORM_EXPORT bool SetOverrideWeightOn (bool);
//! Specify if level material should be used when applying level symbology. @returns true if the new value is different from the current value.
DGNPLATFORM_EXPORT bool SetOverrideMaterialOn (bool);
//! Specify if elements on this level should be displayed. @returns true if the new value is different from the current value.
DGNPLATFORM_EXPORT bool SetDisplay (bool);
//! Specify if elements on this level should be plotted. @returns true if the new value is different from the current value.
DGNPLATFORM_EXPORT bool SetPlot (bool);
//! \em TBD...
DGNPLATFORM_EXPORT bool SetByCell (bool);
//! Specify if elements on this level may be snapped to. @returns true if the new value is different from the current value.
DGNPLATFORM_EXPORT bool SetSnap (bool);
//! Specify if this level is hidden, i.e., should not be shown to the user. @returns true if the new value is different from the current value.
DGNPLATFORM_EXPORT bool SetHidden (bool);
//! Specify access restrictions to be applied to elements on this level
//! @param access can be one of:
//! -- LevelElementAccess::All: Elements on the level have all access. This is the default mode
//! -- LevelElementAccess::Locked: Elements on the level are locked. They can be copied, but cannot be modified or deleted. New elements can be added to the level - once added they are immediately locked. The level can be set active.
//! @returns true if the new value is different from the current value.
DGNPLATFORM_EXPORT bool SetElementAccess (LevelElementAccess access);
//! Specify if elements on this level are frozen. @returns true if the new value is different from the current value.
DGNPLATFORM_EXPORT bool SetFrozen (bool);
//! Specify if elements on this level may be located. @returns true if the new value is different from the current value.
DGNPLATFORM_EXPORT bool SetLocate (bool);

//! Set the level display priority. @returns true if the new value is different from the current value.
DGNPLATFORM_EXPORT bool SetDisplayPriority (Int32);
//! Set the level display transparency. @returns true if the new value is different from the current value.
DGNPLATFORM_EXPORT bool SetTransparency (double);

//! Set the PlotStyle
DGNPLATFORM_EXPORT bool SetPlotStyle (LevelDefinitionPlotStyle const&);

//! Edit the element-related properties defined by this level
DGNPLATFORM_EXPORT void EditElementRelatedProperties (PropertyContextR context);

/// @}

}; // EditLevelHandle



//=======================================================================================
//!
//! A level cache holds level definitions. This is the abstract base class for all types.
//!
//! Supports quick look-up of level definitions by LevelId, LevelCode, and name.
//!
//! A LevelCache may or may not be persistent. This depends on the subclass.
//!
//! \em WARNING: LevelCache is an abstract base class. Make sure you know what purpose
//! the level cache serves before you search it. For example, a reference attachment override level cache
//! will contain only a few levels, while a file level cache will contain all levels used in a file.
//!
//! <h3>Collections and Iterators</h3>
//! See \ref LevelCacheCodeSamples_Iterate
//!
//! <h3>Example Code</h3>
//! See \ref LevelCacheCodeSamples
//!
//  @remark Required library : DgnPlatform<ApiNumber>.lib  i.e. DgnPlatform5.lib
// @bsiclass                                                     Bentley Systems

//=======================================================================================
struct LevelCache : RefCountedBase
{

public:


    //! Look up a level definition by its internal LevelId
    //! Example: \ref LevelCacheCodeSamples_GetElementLevelDefinition
    //! Example: \ref LevelCacheCodeSamples_GetByLevelColor
    //! @param levelId          the level Id to look up
    //! @param includeLibraries if true, search associated level libraries for levelId in addition to this LevelCache.
    //! @returns A handle to the level or an invalid handle if \a levelId was not found. Check the value of LevelHandle::IsValid to detect failure.
    DGNPLATFORM_EXPORT LevelHandle                      GetLevel (LevelId levelId, bool includeLibraries=false) const;

    //! Look up a level definition by its user-defined LevelCode
    //! @param levelCode           The level code to look up
    //! @param includeLibraries if true, search associated level libraries for levelCode in addition to this LevelCache.
    //! @returns A handle to the level or an invalid handle if no level with this code was not found. Check the value of LevelHandle::IsValid to detect failure.
    DGNPLATFORM_EXPORT LevelHandle                      GetLevelByCode (LevelCode levelCode, bool includeLibraries=false) const;

    //! Look up a level definition by its user-defined name
    //! @param levelName           The name to look up
    //! @param includeLibraries if true, search associated level libraries for levelName in addition to this LevelCache.
    //! @returns A handle to the level or an invalid handle if no level by this name was not found. Check the value of LevelHandle::IsValid to detect failure.
    DGNPLATFORM_EXPORT LevelHandle                      GetLevelByName (WCharCP levelName, bool includeLibraries=false) const;

    //! Look up a level definition by its (formatted) display name
    //! @param levelDisplayName  The name to look up
    //! @param includeLibraries if true, search associated level libraries for levelName in addition to this LevelCache.
    //! @returns A handle to the level or an invalid handle if no level by this name was not found. Check the value of LevelHandle::IsValid to detect failure.
    DGNPLATFORM_EXPORT LevelHandle                      GetLevelByDisplayName (WCharCP levelDisplayName, bool includeLibraries=false) const;

    typedef LevelHandle const_iterator;
    typedef const_iterator iterator;    //!< only const iteration is possible

    //! Returns an iterator over the levels in this cache.
    //! Example: \ref LevelCacheCodeSamples_Iterate
    DGNPLATFORM_EXPORT LevelHandle                      begin () const;
    //! Returns an iterator marking the end of the levels in this cache.
    //! Example: \ref LevelCacheCodeSamples_Iterate
    DGNPLATFORM_EXPORT LevelHandle                      end () const;

    //! Returns the count of level definitions in this cache.
    //! @param[in] includeLibraries If true, get the total count of levels in the level cache and all of its associated level libraries.
    DGNPLATFORM_EXPORT size_t                           GetLevelCount (bool includeLibraries=false) const;

    //! Computes and returns the highest LevelId used in this cache.
    DGNPLATFORM_EXPORT LevelId                          ComputeHighestUsedLevelId () const;
    //! Computes and returns the highest LevelCode used in this cache.
    DGNPLATFORM_EXPORT LevelCode                        ComputeHighestUsedLevelCode () const;
    //! Get the lowest available LevelId in this level cache
    DGNPLATFORM_EXPORT LevelId                          GetLowestAvailableLevelId () const;
    //! Get the lowest available LevelCode in this level cache
    DGNPLATFORM_EXPORT LevelCode                        GetLowestAvailableLevelCode () const;

    //! Gets the time that file that defines the levels in this cache was last modified.
    //! @param[out] timeStampOut    The file's last-modified timestamp in milliseconds since 1970
    //! @return non-zero error status if this level cache is new and has never been written to file.
    DGNPLATFORM_EXPORT LevelCacheErrorCode              GetLastModifiedTimeForLevelDefinitions (UInt32& timeStampOut) const;

    //! Query if the level cache is read-only
    DGNPLATFORM_EXPORT bool                             IsReadOnly () const;

    //! Compare this level cache to \a rhs.
    //! @param rhs                          the other level cache
    //! @param compareElementIds            If true, the ElementIds of the levels are also compared. Defaults to false.
    //! @param compareAttributesMask        Optional bit mask to specify what level properties to compare. Defaults to all properties.
    //! @return true if the levels in this level match the levels in \a rhs.
    //! @remarks This function considers only level data, not any data or flags that may be associated with the level cache itself.
    DGNPLATFORM_EXPORT bool                             EqualLevelData (LevelCache const& rhs, bool compareElementIds = false, BitMaskCP compareAttributesMask = NULL) const;

    //! Check if the "hidden" state of individual levels is persisted when converting to element.
    //! @return true if the hidden state of individual levels is persisted
    DGNPLATFORM_EXPORT bool                             IsHiddenStatePersisted () const;
    //! Set if the "hidden" state of individual levels is persisted when converting to element.
    //! @param persistStateIn               If true, then the "hidden" state of individual levels is persisted
    DGNPLATFORM_EXPORT void                             SetPersistHiddenState (bool persistStateIn);

};


//=======================================================================================
//! Controls Change writing for LevelCache EditProperties method.
//=======================================================================================
enum class  EditPropertiesChangeWrite
    {
    Always              = 0,    //!< Write changes to the file regardless of whether there were  pre-existing changes.
    IfNewChangesOnly    = 1,    //!< Write changes to the file only if there were previously no unsaved changes.
    Never               = 2,    //!< Don't write changes to the file.
    };

//=======================================================================================
//! Abstract base class for persistent LevelCaches that support Write.
//! NB:
//! All levels in a level cache must come from the same DGN file. (The color codes and style IDs stored
//! in the level cache refer to objects in the DGN file.)
//  @remark Required library : DgnPlatform<ApiNumber>.lib  i.e. DgnPlatform5.lib
// @bsiclass                                                     Bentley Systems

//=======================================================================================
struct PersistentLevelCache : LevelCache
{
public:
    //! Write this level cache, including all of the level definitions that it contains, to the file that contains it.
    //! @return non-zero error status if the file update failed.
    DGNPLATFORM_EXPORT LevelCacheErrorCode Write ();

    //! Returns true if this level cache or any level definition in it has been modified since it was created or since the last call to Write.
    DGNPLATFORM_EXPORT bool IsDirty () const;

    //! Edit level properties in the level cache
    //! @remarks If the levelCache is changed by the operation, the changes are written according to the writeChanges argument.
    //! @param context      The property-editing context to apply to all levels in the LevelCache
    //! @param writeChanges Controls whether changes are written to the file.
    DGNPLATFORM_EXPORT void                                 EditProperties (PropertyContextR context, EditPropertiesChangeWrite writeChanges);


};


//! Smart pointer wrapper for FileLevelCache.
typedef RefCountedPtr<FileLevelCache>   FileLevelCachePtr;
typedef bvector<FileLevelCachePtr>      T_LevelCachePtrVector;

//=======================================================================================
//! A FileLevelCache contains the level definitions for a file.
//!
//! A FileLevelCache always owned by a DgnFile.
//! The lifetime of a FileLevelCache is contained within the lifetime of its file. A file might
//! delete and recreate its FileLevelCache during its lifetime.
//!
//! A DgnFile always has a FileLevelCache. See DgnFile::GetLevelCacheR, DgnModel::GetFileLevelCacheP.
//!
//! A FileLevelCache is always persistent. It may be marked as read-only. See FileLevelCache::IsReadOnly.
//!
//! A FileLevelCache contains the definition of every level that can be used in a file. The color codes
//! and style IDs stored in these level definitions refer to color tables, color books, and style tables in that file.
//!
//! A FileLevelCache is where you create new levels. You can also copy levels into a FileLevelCache.
//!
//! <h3>Collections and Iterators</h3>
//! See \ref LevelCacheCodeSamples_Iterate
//!
//! <h3>Example Code</h3>
//! See \ref LevelCacheCodeSamples
//!
//  @remark Required library : DgnPlatform<ApiNumber>.lib  i.e. DgnPlatform5.lib
// @bsiclass                                                     Bentley Systems

//=======================================================================================
struct          FileLevelCache : PersistentLevelCache
{
private:
    virtual void MakeClassAbstract() = 0;
public:

public:
    DGNPLATFORM_EXPORT static RefCountedPtr<FileLevelCache> Create (DgnFileR file);

    //! Get the DgnFile that owns this level cache.
    DGNPLATFORM_EXPORT DgnFileR                         GetDgnFileR () const;

    //! Create a level in this level cache.
    //! @param      levelName          Name to be assigned to the new level. Pass NULL if level is to have no name.
    //! @param      levelCode          Code to be assigned to the new level. Pass LEVEL_NULL_CODE to have this LevelCache generate and assign a LevelCode to the new level.
    //! @param      idIn            Internal LevelId to be assigned to the new level. Pass LEVEL_NULL_ID to have this LevelCache generate and assign a LevelId to the new level.
    //! @returns A handle to the newly created level or an invalid handle if creation failed.
    //!         To detect failure, call the handle's LevelDefinition::IsValid method. If that returns false, call LevelDefinition::GetStatus method and check for:
    //!     \li LevelCacheErrorCode::LevelIdIsDuplicate - \a i is a LevelId already used by a level in this cache
    //!     \li LevelCacheErrorCode::LevelNameIsDuplicate - \a n is a name that is already used by a level in this cache
    //!     \li LevelCacheErrorCode::LevelNameIsInvalid - \a n contains invalid characters. See LevelUtils::ContainsInvalidCharacters
    //!     \li LevelCacheErrorCode::LevelCodeIsDuplicate - \a c is a LevelCode already used by a level in this cache
    // Invokes _OnCreateLevel, _OnUseLevelId, and _UseNextLevelCode
    DGNPLATFORM_EXPORT EditLevelHandle                  CreateLevel (WCharCP levelName, LevelCode levelCode, LevelId idIn);

    //! Create the default level using the default level name and LEVEL_DEFAULT_LEVEL_CODE and LEVEL_DEFAULT_LEVEL_ID
    DGNPLATFORM_EXPORT EditLevelHandle                  CreateDefaultLevel ();

    //! Copy the specified level from another file into the level cache of this file.
    //! @remarks This function is useful for copying a level from one file into another. It cannot be used to copy a level within a file.
    //! To copy a level with control over name, code, and what data is copied, use CreateLevel and EditLevelHandle::CopyLevelData
    //! @remarks If successful, this function creates a new level in the level cache of this file and then copies the data from the source level.
    //! The new level is \em always assigned a new internal LevelId. The source level's LevelId is never copied.
    //! @returns A handle to the newly created level or an invalid handle if creation or copying failed.
    //!         To detect failure, call the handle's LevelDefinition::IsValid method. If that returns false, call LevelDefinition::GetStatus method and check for:
    //!     \li LevelCacheErrorCode::LevelNameIsDuplicate - the name of the source level is already used by an existing level in this cache
    //!     \li LevelCacheErrorCode::LevelCodeIsDuplicate - the LevelCode of the source level is already used by a level in this cache
    //!     \li LevelCacheErrorCode::CannotFindLevel - the source level is invalid
    //! @param          source                      The source level to copy.
    //! @param          replaceDuplicateLevelCode   [optional] if true, and the level code is a duplicate, assign a new level code.
    //! @param
    // Invokes _OnCopyLevel after copying the level
    DGNPLATFORM_EXPORT EditLevelHandle                  CopyLevel (LevelHandle const& source, bool replaceDuplicateLevelCode=false);

    //! Remove a level from the cache
    //! Example: \ref LevelCacheCodeSamples_RemoveLevel
    //! @param[out] next  The next level in the level cache. Useful when deleting levels while iterating a level cache. Pass NULL if you do not want to know the next level.
    //! @param[in]  level The level to remove. <em>Becomes invalid.</em>
    //! @returns Non-zero error status if the level cannot be removed:
    //! @remarks If this function succeeds, the input level becomes invalid.
    //!     \li LevelCacheErrorCode::CannotFindLevel - The specified level is invalid or is not from this level cache.
    //!     \li LevelCacheErrorCode::TableIsReadOnly - The level table is read-only
    //!     \li LevelCacheErrorCode::LevelIsReadOnly - The level itself is read-only
    //!     \li LevelCacheErrorCode::LevelIsUsed - The level is in use
    // Invokes _OnRemoveLevel before removing the level. The level is not removed if _OnRemoveLevel returns non-zero.
    DGNPLATFORM_EXPORT LevelCacheErrorCode              RemoveLevel (LevelHandle* next, LevelHandle& level);

    //! Mark a level read-only or read-write
    DGNPLATFORM_EXPORT bool                             SetLevelReadOnly (LevelHandle&, bool);

    //! Get the highest used LevelId in this level cache.
    //! @param[in] includeLibraries If true, get the highest use LevelId in the level cache and all of its associated level libraries.
    DGNPLATFORM_EXPORT LevelId                          GetHighestUsedLevelId (bool includeLibraries=false) const;

    //! Get the highest used LevelCode in this level cache
    DGNPLATFORM_EXPORT LevelCode                        GetHighestUsedLevelCode() const;

    //! Test if any element in specified model-ref is on the specified level.
    //! See \ref LevelCache_Usage
    DGNPLATFORM_EXPORT bool                             IsLevelUsedInModel (DgnModelRefP, LevelId) const;
    //! Test if any element in file of "this" FileLevelCache is on the specified level
    //! See \ref LevelCache_Usage
    DGNPLATFORM_EXPORT bool                             IsLevelUsedInFile (LevelId) const;

    //! Get a vector of Level Libraries associated with this LevelCache.
    DGNPLATFORM_EXPORT T_LevelCachePtrVector*           GetLibraries () const;

};

//=======================================================================================
//! Provides access to levels used by elements in a referenced model. Contains level definitions
//! that override levels in the referenced file.
//!
//! A DgnAttachmentLevelCache is always owned by a DgnAttachment. Call
//! DgnAttachment::GetAttachmentLevelCache to get the override level cache for a reference.
//!
//! DgnModelRef::GetLevelCacheR calls GetAttachmentLevelCache when the modelref is a DgnAttachment.
//!
//! A LevelHandle obtained from a DgnAttachmentLevelCache may refer to
//! a level override defined in the DgnAttachmentLevelCache itself, or it may
//! refer to a level defined in the referenced model. Moreover, level overrides stored in a
//! DgnAttachmentLevelCache have the ability to refer to color codes and style IDs
//! that are actually defined in the reference file. See for LevelHandle for how to
//! query for the file that defines the color or style of a level.
//!
//! DgnAttachmentLevelCache supports the concept of "synchronization" with the FileLevelCache from the DgnAttachment's file.
//!
//! <h3>Collections and Iterators</h3>
//! See \ref LevelCacheCodeSamples_Iterate
//!
//! <h3>Example Code</h3>
//! See \ref LevelCacheCodeSamples
//!
//  @remark Required library : DgnPlatform<ApiNumber>.lib  i.e. DgnPlatform5.lib
// @bsiclass                                                     Bentley Systems

//=======================================================================================
struct          DgnAttachmentLevelCache : PersistentLevelCache
{
private:
    virtual void MakeClassAbstract() = 0;
public:

public:

    //! Get the reference attachment that owns this level cache
    DGNPLATFORM_EXPORT DgnAttachmentR                   GetDgnAttachmentR () const;

};



//  @remark Required library : DgnPlatform<ApiNumber>.lib  i.e. DgnPlatform5.lib
// @bsiclass                                                     Bentley Systems

struct          LevelUtils
{
//! Query if the specified string is not a valid level name
DGNPLATFORM_EXPORT static bool ContainsInvalidCharacters (WCharCP);
//! Modify the specified string so that it is a valid level name
//! @DotNetMethodExclude
DGNPLATFORM_EXPORT static void ReplaceInvalidCharacters (WCharP, WChar);
//! Modify the specified string so that it is a valid level name
//! @DotNetMethodExclude
DGNPLATFORM_EXPORT static void ReplaceInvalidCharacters (WStringR, WChar);

}; // LevelUtils


/// @}  -- LevelCache

/**
@addtogroup LevelCacheCodeSamples

\li \ref LevelCacheCodeSamples_GetElementLevelDefinition
\li \ref LevelCacheCodeSamples_Iterate
\li \ref LevelCacheCodeSamples_GetByLevelColor
\li \ref LevelCacheCodeSamples_CreateLevel
\li \ref LevelCacheCodeSamples_RemoveLevel

@{
\section LevelCacheCodeSamples_GetElementLevelDefinition Get level definition data for an element

If you have an ElementRef, information about its level in the file in which it is stored is always available
by getting the level from the FileLevelCache of its DgnModel.

\code
    #include <DgnPlatform/LevelCache.h>

    void getElementLevel (ElementRefP elementRef)
        {
        // Since we have an elementRef from the DgnModel, we know that FileLevelCache is never NULL.
        FileLevelCache* fileLevelCache = elementRef->GetDgnModelP()->GetFileLevelCacheP ();

        LevelHandle level = levelCache.GetLevel (elementRef->GetLevel ());

        printf ("level name is %S\n", level.GetName ());
        }
\endcode

    If you are or may be viewing the element through a DgnAttachment, then ask the model reference for the level.

\code
    void getElementLevelPossiblyOverride (ElementRefP ref, DgnModelRefP modelRef)
        {
        //  If modelRef is a DgnAttachment, its GetLevelCacheR method will return a DgnAttachmentLevelCache.
        //  If modelRef is an ordinary DgnModel, its GetLevelCache method will return the FileLevelCache of the file that contains it.

        LevelCache& levelCache = modelRef->GetLevelCacheR ();

        LevelHandle level = levelCache.GetLevel (ref->GetLevel ());

        printf ("by-level weight is %d\n", level.GetByLevelWeight());
        }

    \endcode
    \see Bentley::DgnModelRef::GetFileLevelCacheP,
    Bentley::DgnModelRef::GetLevelCacheR,
    Bentley::DgnPlatform::LevelCache::GetLevel,
    Bentley::DgnPlatform::LevelHandle::GetByLevelWeight,
    Bentley::DgnPlatform::LevelHandle::GetName

\section LevelCacheCodeSamples_Iterate Iterate levels in a level cache

    A LevelCache is an STL collection of LevelDefinitionHandles. You can iterate it
    using standard iteration syntax:

    \code
    void iterateLevelsStd (DgnFileR file)
        {
        LevelCache& levelCache = file.GetLevelCacheR();
        for (LevelHandle level (levelCache.begin()); level != levelCache.end(); ++level)
            {
            if (level.GetDisplay ())
                ... do something ...
            }
        }
    \endcode

    You can also use Microsoft enhanced iteration syntax:

    \code
    void iterateLevelsEnhanced (DgnFileR file)
        {
        for each (LevelHandle const& level in file.GetLevelCacheR())
            {
            if (level.GetDisplay ())
                ... do something ...
            }
        }
    \endcode

    The most convenient way to get read-write access to all levels in a cache is to iterate using LevelDefinitionEditHandles.
    Note that you cannot use enhanced for-each iteration syntax in this case.

    \code
    void iterateAndEditLevels (DgnFileR file)
        {
        // In this example, increase the by-level weight of each level by 1
        LevelCache& levelCache = file.GetLevelCacheR();
        for (EditLevelHandle level (levelCache.begin()); level != levelCache.end(); ++level)
            {
            level.SetByLevelWeight (level.GetByLevelWeight() + 1);
            ...
            }
        levelCache.Write ();    // don't forget to save your changes
        }
    \endcode

    \see Bentley::DgnPlatform::LevelCache::begin,
    Bentley::DgnPlatform::LevelHandle,
    Bentley::DgnPlatform::EditLevelHandle,
    Bentley::DgnPlatform::FileLevelCache::Write

\section LevelCacheCodeSamples_GetByLevelColor Get the by-level color for an element
\code
    #include <DgnPlatform/LevelCache.h>

    USING_NAMESPACE_BENTLEY_MSTNPLATFORM

    void getByLevelColor (ElementHandleCR eh)
        {
        //  If eh has been set up with a modelref, then it might refer to a DgnAttachment. If so, its GetLevelCacheR
        //  method will return a DgnAttachmentLevelCache.
        //  Otherwise, get the FileLevelCache of the file that contains the element.
        LevelCache& levelCache = (eh.GetDgnModelRef() != NULL)? eh.GetDgnModelRef()->GetLevelCacheR(): eh.GetDgnModelP()->GetLevelCacheR ();

        LevelHandle level = levelCache.GetLevel (eh.GetElementCP()->ehdr.level);

        //  Get the by-level color
        LevelDefinitionColor byLevelColor = level.GetByLevelColor ();

        //  A color may be an RGB value, a color book entry, or an entry in the definition file's color table.
        IntColorDef colorDef;
        DgnColorMap::ExtractElementColorInfo (&colorDef, NULL, NULL, NULL, NULL, byLevelColor.GetCode(), byLevelColor.GetDefinitionFile());
        printf ("RBG color is %x\n", colorDef.m_rgb);
        }
    \endcode
    \see Bentley::DgnModelRef::GetLevelCacheR
    \see Bentley::DgnPlatform::LevelHandle::GetByLevelColor

\section LevelCacheCodeSamples_CreateLevel Create a level
\code
    #include <DgnPlatform/LevelCache.h>

    USING_NAMESPACE_BENTLEY_MSTNPLATFORM

    void createLevel (DgnFileR file)
        {
        // You can create levels only in a FileLevelCache.
        FileLevelCache& levelCache = file.GetLevelCacheR ();

        EditLevelHandle level = levelCache.CreateLevel (L"newName");
        if (!level.IsValid ())
            {
            // "newName" might already be in use. Check level.GetStatus() for the value LevelCacheErrorCode::LevelNameIsDuplicate.
            return;
            }

        //  You can set up the level's attributes before writing it to file.
        level.SetByLevelWeight (5);

        //  Save the new level.
        if (levelCache.Write () != LevelCacheErrorCode::None)
            {
            //  levelCache might be read-only, or the file might be read-only.
            }
        }
    \endcode
    \see Bentley::DgnModelRef::GetLevelCacheR
    \see Bentley::DgnPlatform::FileLevelCache::CreateLevel
    \see Bentley::DgnPlatform::FileLevelCache::Write
    \see Bentley::DgnPlatform::EditLevelHandle
    \see \ref LevelCacheCodeSamples_CreateLevel

\section LevelCacheCodeSamples_RemoveLevel Remove a level

    To remove a level when you know its LevelId, LevelCode, or name:

    \code
    void removeThisLevel (DgnFileR file, LevelId levelId)
        {
        LevelCache& levelCache = file.GetLevelCacheR();
        LevelHandle level = levelCache.GetLevel (levelId);
        //  You could check that level.IsValid() == true, or you could just go on to the next step:
        if (levelCache.RemoveLevel (level) != LevelCacheErrorCode::None)
            // level was not found.
        levelCache.Write ();    // don't forget to save your change
        }
    \endcode

    To remove a level while iterating:

    \code
    void removeLevelWhileIterating (DgnFileR file)
        {
        LevelCache& levelCache = file.GetLevelCacheR();
        for (EditLevelHandle level (levelCache.begin()); level != levelCache.end(); )
            {
            if ( ... this is the level to remove ... )
                levelCache.RemoveLevel (&level, level);  // modifies level to point to the next level in the cache
            else
                ++level;
            }
        }
    \endcode

    \see Bentley::DgnPlatform::LevelCache::begin
    \see Bentley::DgnPlatform::LevelCache::GetLevel
    \see Bentley::DgnPlatform::FileLevelCache::RemoveLevel
    \see Bentley::DgnPlatform::FileLevelCache::Write
    \see Bentley::DgnPlatform::LevelHandle

@}
*/

END_BENTLEY_DGNPLATFORM_NAMESPACE

